// Realtime Pipe - TAZ
// Version 2 edited by Scott Lange
// 

@warnings
@version 2.6
@script modeler
@name TAZ_Pipe_v2

/*
v2.4 Lots of changes
    made it compatable with Lightwave 7.x
    add support for custom mesh
        This changed a good bunch of code
    added offset
    added flipped normals
    
v2.3.1
    small fix in the way bevel ends works
v2.3
    add better subpatch support for endes
    added weights
    added beveled ends
v2.2 fixed bugs
    added subpatch
v2.1 fixed two bugs
    points out of order in a curve.
    running out of undos leaving behind garbage 
    Added support for UVs
    Added support for copying surfaces from curves
v2 added support for selection and muliple curves
added support for automative segments
and more flexable layers(will creat geo. in first empty layer 
*/

mode = 1;
f=0;
ft=0;
c0;
c1;
c2;
c3;
c4;
c5;
c6;
c11;
c12;
c7;
c8;
c9;
c10;
c13;
makeUV = 1;
subpatch = 1;
weightmap = 1;
bevends = 1;

pntPos;
segments=10;
dRadius=5;
offsetX = 0;
offsetY = 0;
nSides=4;
dRotate=0;
norm;
avnorm;
rot;
UserGeom=0;
curvesegments;
currentlyr;
segments;
data;
empty1;
empty2;
empty3;
empty4;
totalpolys;

orient;
interp;
CurveID;
PolygonID;
anypolys;
flip;
meshPcount;

main
{

    anyCurves = 0;

//== find layers selected
    currentlyr = lyrfg();
   
//== find first open layer
    empty = GetEmptyLayers(1);
    empty1 = empty[1];
    empty2 = empty[2];
    empty3 = empty[3];
    empty4 = empty[4];
   
    selmode(USER);

    totalpolys = polycount();
	curvesegments = pointcount();
    
    editbegin();
    foreach(poly,polygons)
    {
        
        if(poly.isCurve())
        {
            anyCurves = poly.isCurve();            
            CurveID += poly;
        }
        else
        {
            UserGeom = 1;
            PolygonID += poly;
        }
        
    }
    editend();

    if(anyCurves == 0)
        error("No curves!");
    if(UserGeom == 1)
    {
        makeUV = 0;
        subpatch = 0;
        weightmap = 0;
        bevends = 0;
    }    
    tubes();
}

tubes
{
   
   	reqbegin("Taz/SR Pipes");
	c0 = ctlchoice("Mode",mode,@"Extrude","Clone"@);
	c1 = ctlcheckbox("Oriented",1);
	c2 = ctlchoice("Interpolation",3,@"Knots","Lengths","Curve Segments"@);
	c3 = ctlminislider("Segments",segments,1,100000);
	c4 = ctlminislider("Radius",dRadius,1,1000);
	if(UserGeom == 0)
		c5 = ctlminislider("Sides",nSides,3,32);
	c6 = ctlminislider("Rotate",dRotate,0,359);
	c11 = ctlminislider("Offset X",offsetX,-1000000,1000000);
	c12 = ctlminislider("Offset Y",offsetY,-1000000,1000000);
	c7 = ctlcheckbox("Make UVs",makeUV);
	c8 = ctlcheckbox("SubPatch",subpatch);
	c9 = ctlcheckbox("WeightMap",weightmap);
	c10 = ctlcheckbox("Bevel ends",bevends);
	c13 = ctlcheckbox("Flip Normals",0);
	
	if(ft==0)
	{
		MTtubes(1);
		ft=1;
	}
	
	ctlrefresh(c0,"MTtubes");
	ctlrefresh(c1,"MTtubes");
	ctlrefresh(c2,"MTtubes");
	ctlrefresh(c3,"MTtubes");
	ctlrefresh(c4,"MTtubes");
	if(UserGeom == 0)
	    ctlrefresh(c5,"MTtubes");
	ctlrefresh(c6,"MTtubes");
	ctlrefresh(c7,"MTtubes");
	ctlrefresh(c8,"MTtubes");
	ctlrefresh(c9,"MTtubes");
	ctlrefresh(c10,"MTtubes");
	ctlrefresh(c11,"MTtubes");
	ctlrefresh(c12,"MTtubes");
	ctlrefresh(c13,"MTtubes");

	if(!reqpost())
	{
		if(f==0)
		{
			return;
		}
		else
		{
	    /*
		    if(15*totalpolys[1] < 128)
		    {
		        for(u=1;u<=15*totalpolys[1];u++)
		            undo();
		        undo();
		        undo();
            }
            else
            {
                lyrsetfg(empty1,empty2,empty3);
                delete();
            }
		//if(UserGeom==0)
			//undo();
	    */
	    lyrsetfg(empty1,empty2,empty3);
        delete();
			lyrsetfg(currentlyr);
			return;
		}
	}

    reqend();
    
}

// multi curve railexstrude
MTtubes: val
{
    vmap;

	if(f==1)
	{
	    /*
		    if(15*totalpolys[1] < 128)
		    {
		        for(u=1;u<=15*totalpolys[1];u++)
		            undo();
		        undo();
		        undo();
            }
            else
            {
                lyrsetfg(empty1,empty2,empty3);
                delete();
            }
		//if(UserGeom==0)
			//undo();
	    */
	    lyrsetfg(empty1,empty2,empty3);
        delete();
	}
			
	f=1;
	mode = getvalue(c0);
	orient = getvalue(c1);
	interp = getvalue(c2);
	segments = getvalue(c3);
    dRadius = getvalue(c4) / 100;
    if(UserGeom == 0)
	    nSides = getvalue(c5);
	makeUVs = getvalue(c7);
	subpatch = getvalue(c8);
	weightmap = getvalue(c9);
	bevends = getvalue(c10);
	offsetX = getvalue(c11) / 100;
	offsetY = getvalue(c12) / 100;
	flip = getvalue(c13);

//== select the uv map or make one   
    vmap = VMap(VMTEXTURE,0);
    wvmap = VMap(VMWEIGHT,0);
    
    if(vmap == nil && makeUVs == 1)
        info("Taz_PipeUV map will be created");
    
    if(wvmap == nil)
        info("Taz_PipeWeight map will be created");


//== loop though each curve    
    for(i=1;i<=CurveID.size();i++)
    {
        //code clean up
        
    	lyrsetfg(currentlyr);
        selpolygon(CLEAR);

//====== Clear variable        
        curvesegments = nil;
        for(n=1;n<=j.size();n++)
            j[n] = nil;
        n = nil;
        for(n=1;n<=p.size();n++)
            p[n] = nil;
        n = nil;
        for(n=1;n<=crvPnts.size();n++)
            crvPnts[n] = nil;
        for(n=1;n<=pntPos.size();n++)
            pntPos[n] = nil;
        n = nil;
        j = nil;
        p = nil;
        crvPnts = nil;
        pntPos = nil;
        pnt2 = nil;
        pc = nil;
        
        data = nil;
        
        Useg = nil;
        U = nil;
        V = nil;
        Uspace = nil;
        Vspace = nil;
        
             
//====== select poly and start railing

        selpolygon(SET,POLYID,CurveID[i]);
        copy();    
//====== select first open layer
 
        lyrsetfg(empty3);
        paste();
        
        pc = editbegin();
        
        data = polyinfo(polygons[1]);

        for(j=1;j<data.size();j++)
        {
            crvPnts+=data[j+1];
        }

        foreach(p,crvPnts)
		    pntPos += pointinfo(p);
		      	
    	// first curve
    	p[2] = pntPos[1];
    	p[3] = pntPos[2];
    	p[4] = pntPos[3];
    	p[1] = (p[2] - p[3]) * 3.0 + p[4];
    	
    	
    	pnt2 = Curve(0.01, p[1], p[2], p[3], p[4]);
    	norm = normalize(pnt2 - pntPos[1]);
    
    	editend();
    	
    	curvesegments = pointcount();
 
//====== Perform rail exstrude 

        if(UserGeom == 0)
            dodisc();
        else
            domesh();

        lyrsetfg(empty2);
        lyrsetbg(empty3);

          	
    	if(orient==1) os="o";
    	else os="s";
    
    	if(interp==1)
    	{
		    if(mode==1)
    			railextrude(segments,KNOTS,os);
		    if(mode==2)
    			railclone(segments,KNOTS,os);
        }
        if(interp==2)
    	{
		    if(mode==1)
    		    railextrude(segments,LENGTHS,os);
		    if(mode==2)
    		    railclone(segments,LENGTHS,os);
        }
    	if(interp==3)
    	{
            if(mode==1)
    			railextrude(curvesegments - 1,KNOTS,os);
		    if(mode==2)
    			railclone(curvesegments - 1,KNOTS,os);
    	}

//====== apply surfaces from curves
    	changesurface(data[1]);

   	
    	if(makeUVs == 1)
    	{
    	    Useg = 0;
            U = 0;
            V = 0;

            if(UserGeom == 1)
                nSides = meshPcount;
                
            Uspace = 1/(nSides.size());
            
            if(interp==3)
                Vspace = 1/(curvesegments.size() - 1);
            else
                Vspace = 1/segments.size();

            editbegin();
            
            if(vmap == nil && makeUVs == 1)
                vmap = VMap(VMTEXTURE,"Taz_PipeUV",2);
            
//========== loop though all points
            foreach(uv,points)
            {
                vmap.setValue(uv,@U,V@);
                if(Useg == (nSides.size() - 1))
                {
                    Useg = 0;
                    U = 0;
                    V = V+Vspace;
                }
                else
                {
                    Useg++;
                    U = U+Uspace;
                }                           
            }        
            editend();
        } 
        
        if(weightmap == 1)
        {
            wseg = 0;
            wval = 1;
            
            if(UserGeom == 1)
                nSides = meshPcount;
                
            Uspace = 1/(nSides.size());
            
            if(interp==3)
                Vspace = 1/curvesegments.size();
            else
                Vspace = 1/segments.size();
                
            editbegin();
            
            if(wvmap == nil)
                wvmap = VMap(VMWEIGHT,"Taz_PipeWeight",2);
            
//========== loop though all points

            foreach(w,points)
            {
                wvmap.setValue(w,wval);
                if(wseg == (nSides.size()))
                {
                    wseg = 0;
                    wval = wval-Vspace;
                }
                else
                {
                    wseg++;
                }
                          
            }        
            editend();
        }
        
    	cut();
    	lyrsetfg(empty1);
    	paste();
    	lyrsetfg(empty2,empty3);
    	delete();
    }
    lyrsetfg(empty1);
    lyrsetbg(currentlyr);
    selpolygon(CLEAR);
    if(bevends == 1)
    {
        if(nSides > 4 || UserGeom == 1)
        {
            selpolygon(SET,NVGT,4);
            bevel(0,.01);
            bevel(.01,0);
        }
    }

    if(subpatch == 1)
    {
        if(hostVersion() >= 8)
        {
            if(nSides > 4 || UserGeom == 1)
            {
                selpolygon(CLEAR);
                selpolygon(SET,NVGT,4);
                cmdseq("Create Pole");
                //cmdseq("Fix Poles");
            }
        }
        selpolygon(SET,NVLT,5);
        togglepatches();
    }
    selpolygon(CLEAR);
}

GetEmptyLayers:n
{
    var empty;
    empty=lyrempty();
    if(empty==nil || size(empty) < n){
        error("Out of empty layers.");
    }
    empty.sortA();
    return empty;
}

Curve: v,p1, p2, p3, p4
{
	var b1, b2, b3, b4;
	b1 = p2;
	b4 = p3;
	var l1, l2, r;
	l2 = vmag(p2 - p3);
	l1 = vmag(p1 - p2);
	if(l1 != 0.0)
	{
		r = l2 / l1;
		b2 = p2 + ((p3 - p2) - (p1 - p2) * r) / ((r + 1) * 3);
	}
	else
	{
		b2 = p2;
	}
	l1 = vmag(p3 - p4);
	if(l1 != 0.0)
	{
		r = l2/l1;
		b3 = p3 + ((p2 - p3) - (p4 - p3) * r) / ((r + 1) * 3);
	}
	else
	{
		b3 = p3;
	}
    vec2 = BezierFunc(v, b1, b2, b3, b4);

	return(vec2);
}

BezierFunc: t, p1, p2, p3, p4
{
	var k1, k2, k3, k4, t1;
	t1 = 1 - t;
	k1 = t1 * t1;
	k4 = t * t;
	k2 = 3 * t * k1;
	k3 = 3 * t1 * k4;
	k1 *= t1;
	k4 *= t;
	return (p1 * k1 + p2 * k2 + p3 * k3 + p4 * k4);
}


convertForwardVector2HPB: n		// n is a normalized vector
{
	// credit: Ernie Wright
	// takes a normalized vector and computes for its Euler coordinates: heading, pitch, and bank

	rotateV = <0,0,0>;
	rotateV.y = asin( -n.y );
	if ( 1.0 - abs( n.y ) > 0.0000000001 )
	{
		rotateV.x = acos( n.z / cos( rotateV.y ));
		if ( n.x < 0.0 )
			rotateV.x = 2 * PI - rotateV.x;
	}
	else
		rotateV.x = 0.0;

	rotateV.x = deg(rotateV.x);
	if(rotateV.x > 180)
		rotateV.x = rotateV.x - 360;

	rotateV.y = deg(rotateV.y);
	return(rotateV);
}

dodisc
{
	dRotate=getvalue(c6);
	
    lyrsetfg(empty2);
	makedisc(dRadius,0,0,"Z",nSides,1,<0,0,0>);
    if(flip == 0)
        flip();

    move(offsetX,offsetY,0);

	rot = -convertForwardVector2HPB(norm);

//== find center

    Center = <offsetX,offsetY,0>;
    
    rotate(dRotate,"Z",Center);
        	
	rotate(rot.y,"X",Center);
	rotate(rot.x,"Y",Center);
	
	move(pntPos[1]);

}

domesh
{
    dRotate=getvalue(c6);
    
    for(i=1;i<=PolygonID.size();i++)
    {
        lyrsetfg(currentlyr);
        selpolygon(CLEAR);
        selpolygon(SET,POLYID,PolygonID[i]);
        copy();
        lyrsetfg(empty2);
        paste();


    }

    meshPcount = pointcount();
    
//== edit polys    
    
    if(flip == 1)
        flip();
    move(offsetX,offsetY,0);

	rot = -convertForwardVector2HPB(norm);

//== find center

    Center = <offsetX,offsetY,0>;

    scale(<dRadius,dRadius,1>,Center);
    rotate(dRotate,"Z",Center);

	rotate(rot.y,"X",<0,0,0>);
	rotate(rot.x,"Y",<0,0,0>);
	    
    move(pntPos[1]);
}